#include <err.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count;) {
    unsigned char *orig_bytes = ((unsigned char*)_data) + byte_offset;
    unsigned char bytes[16];
    memcpy(bytes, orig_bytes, 16);
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
    byte_offset += 16;
  }
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int gpu_fd;
static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;

#define PVRSRV_MAX_BRIDGE_OUT_SIZE 0x1000

#define PVRSRV_BRIDGE_SRVCORE 1UL
#define PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAIT 5
typedef struct PVRSRV_BRIDGE_IN_EVENTOBJECTWAIT_TAG
{
	IMG_HANDLE hOSEventKM;
} __packed PVRSRV_BRIDGE_IN_EVENTOBJECTWAIT;
/* Bridge out structure for EventObjectWait */
typedef struct PVRSRV_BRIDGE_OUT_EVENTOBJECTWAIT_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_EVENTOBJECTWAIT;

int main(void) {
  setbuf(stdout, NULL);
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));

  while (1) {
    PVRSRV_BRIDGE_IN_EVENTOBJECTWAIT in = { .hOSEventKM = (void*)0x12345678/*bogus handle*/ };
    char out[PVRSRV_MAX_BRIDGE_OUT_SIZE];
    memset(out, 'E', sizeof(out));
    int res = driver_cmd(PVRSRV_BRIDGE_SRVCORE, PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAIT, &in, out, sizeof(in), sizeof(out));
    printf("ioctl returned %d\n", res);
    hexdump(out, 0x300);
    printf("\n\n\n");
    sleep(1);
  }
}
